CLI_CONFIG = {
        'unit': {
            'options': ['-u'],
            'default': 'main',
            'sub': ['create', 'get', 'set', 'migrate'],
            'help': 'The unit to work with, if empty the unit "main" is used'
            },
        'store': {
            'default': 'file',
            'sub': ['create', 'migrate'],
            'help': 'Choose the storage medium to use',
            },
        'cipher': {
            'options': ['-C'],
            'default': 'fernet',
            'sub': ['create', 'migrate'],
            'help': 'When creating or cipher to use for storage of data for the new unit',
            },
        'seal': {
            'options': ['-S'],
            'default': 'passwd',
            'sub': ['create', 'migrate'],
            'help': 'The type of seal to use to secure the storage interface',
            },
        'seal_raw': {
            'default': None,
            'sub': ['set', 'get'],
            'help': 'DO NOT USE! This option allows you to pass secrets as command line arguments! This should only be used for testing!!',
            },
        'path': {
            'options': ['-p'],
            'default': None,
            'sub': ['set', 'get'],
            'help': 'The path to store or retrive the desired data',
            },
        'string': {
            'options': ['-s'],
            'defalt': None,
            'sub': 'set',
            'help': 'Choose a string to set as the value for the given storage path',
            },
        'file': {
            'options': ['-f'],
            'default': None,
            'sub': 'set',
            'help': 'Choose a local file to set as the value to find for the given path',
            },
        'data_dir': {
            'default': '/var/takara/data',
            'sub': ['create', 'set', 'get', 'migrate'],
            'help': 'The directory to store data specific to the configuration of encrypted units when using local file storage systems',
            },
        'unit_dir': {
            'default': '/var/takara/unit',
            'sub': ['create', 'set', 'get', 'migrate'],
            'help': 'The directory to store encrypted data when using the local file storage system'
            },
        }
CONFIG = {
        'data_dir': {
            'default': '/var/takara/data',
            'help': 'The directory to store data specific to the configuration of encrypted units when using local file storage systems',
            },
        'unit_dir': {
            'default': '/var/takara/unit',
            'help': 'The directory to store encrypted data when using the local file storage system'
            },
        'store': {
            'default': 'file',
            'help': 'Choose the storage medium to use',
            },
    }
GLOBAL = {}
SUBS = {
        'create': {
            'desc': 'Create New Units',
            'help': 'Use create to make new units to store encrypted data',
            },
        'set': {
            'desc': 'Set data to a location',
            'help': 'Use set to change or add a location for a given unit',
            },
        'get': {
            'desc': 'Get data from a location',
            'help': 'Use Get to retrive data from a given location',
            },
        'migrate': {
            'desc': 'Migrate a unit to other stores, seals, and/or ciphers',
            'help': 'Migrate a unit to other stores, seals, and/or ciphers'
            },
        }
DYNE = {
        'takara': ['takara'],
        }



# Import local libs
import takara.exc


__func_alias__ = {'set_': 'set'}


def __init__(hub):
    hub.pop.sub.load_subdirs(hub.takara)
    hub.takara.UNITS = {}
    hub.takara.CRYPT = {}


def cli(hub):
    '''
    Start the direct cli interface for takara
    '''
    hub.pop.conf.integrate(['takara'], cli='takara', loader='yaml', roots=True)
    kw = hub.OPT['takara']
    hub.pop.loop.start(hub.takara.init.start_cli(**kw))


async def start_cli(hub, **kw):
    await hub.takara.init.setup(**kw)
    coro = {'create': hub.takara.init.create,
     'set': hub.takara.init.set,
     'get': hub.takara.init.get}[hub.OPT['_subparser_']](**kw)
    ret = await coro
    if ret and hub.OPT['_subparser_'] == 'create':
        print('The following key(s) have been generated for your encrypted unit:')
        for key in ret.split(':'):
            print(key)
    if ret and hub.OPT['_subparser_'] == 'get':
        print(ret)


async def setup(hub, **kw):
    '''
    Given the store and the kwargs to access the given store
    '''
    store = kw['store']
    await getattr(hub, f'takara.store.{store}.config')(**kw)


async def create(hub, **kw):
    '''
    Set up the environment based on the available data
    '''
    cipher = kw['cipher']
    seal = kw['seal']
    store = kw['store']
    kw['seal_raw'] = await getattr(hub, f'takara.seal.{seal}.gen')(
        seal_raw=kw.get('seal_raw', None),
        cipher=cipher)
    kw['seal_data'] = await getattr(hub, f'takara.seal.{seal}.create')(**kw)
    ret = await getattr(hub, f'takara.store.{store}.create')(**kw)
    if hasattr(hub, f'takara.seal.{seal}.EXPOSE_SEAL_RAW'):
        if getattr(hub, f'takara.seal.{seal}.EXPOSE_SEAL_RAW'):
            return kw['seal_raw']
    return ret


async def unseal(hub, **kw):
    '''
    Unseal the desired unit.
    '''
    unit = kw['unit']
    unit_config = hub.takara.UNITS[unit]
    if 'crypter' in unit_config:
        # Already unsealed, carry on!
        return True
    seal = unit_config['seal']
    cipher = unit_config['cipher']
    seal_data = unit_config['seal_data']
    if kw.get('seal_raw'):
        seal_raw = kw['seal_raw']
    else:
        if hasattr(hub, f'takara.seal.{seal}.get'):
            seal_raw = await getattr(hub, f'takara.seal.{seal}.get')(kw.get('passwd', None))
        else:
            seal_raw = await getattr(hub, f'takara.seal.{seal}.gen')(kw.get('passwd', None))
    if not await getattr(hub, f'takara.seal.{seal}.verify')(seal_raw, seal_data):
        return False
    await getattr(hub, f'takara.cipher.{cipher}.setup')(unit, seal_raw)
    return True


async def set_(hub, **kw):
    '''
    Set a specific value to a specific path
    '''
    if kw['unit'] not in hub.takara.UNITS:
        raise takara.exc.UnitMissingError()
    unit = kw['unit']
    unit_config = hub.takara.UNITS[unit]
    cipher = unit_config['cipher']
    store = unit_config['store']
    # TODO: If we need to add data sources make this a sub
    if kw.get('file'):
        with open(kw['file'], 'rb') as rfh:
            data = rfh.read()
    elif kw.get('string'):
        data = kw['string'].encode('utf-8')
    unsealed = await hub.takara.init.unseal(**kw)
    if not unsealed:
        raise takara.exc.UnsealError('Falied to unseal the treasure, was entry correct?')
    enc = await getattr(hub, f'takara.cipher.{cipher}.encrypt')(unit, data)
    await getattr(hub, f'takara.store.{store}.set')(unit, kw['path'], enc)


async def get(hub, **kw):
    '''
    Get a specific value from a specific path
    '''
    if kw['unit'] not in hub.takara.UNITS:
        raise takara.exc.UnitMissingError()
    unit = kw['unit']
    unit_config = hub.takara.UNITS[unit]
    cipher = unit_config['cipher']
    store = unit_config['store']
    unsealed = await hub.takara.init.unseal(**kw)
    if not unsealed:
        raise takara.exc.UnsealError('Falied to unseal the treasure, was entry correct?')
    enc = await getattr(hub, f'takara.store.{store}.get')(unit, kw['path'])
    ret = await getattr(hub, f'takara.cipher.{cipher}.decrypt')(unit, enc)
    return ret.decode()



'''
Implement the fernet algorithm from pycryptography
'''
# Import python libs
import hashlib
import base64
# Import third party libs
import cryptography.fernet


def gen(hub):
    '''
    Generate a raw key to use for fernet. This is used by seals that require
    a key to be generated by the cipher, like shamir and not by seals that
    derive the key weakly from use input, like passwd
    '''
    return cryptography.fernet.Fernet.generate_key()


async def setup(hub, unit, seal_raw):
    '''
    Using the seal data derive a key
    '''
    seal_raw = seal_raw.encode()
    seal = hub.takara.UNITS[unit]['seal']
    if hasattr(hub, f'takara.seal.{seal}.ENCODE_RAW'):
        if getattr(hub, f'takara.seal.{seal}.ENCODE_RAW'):
            key = base64.urlsafe_b64encode(hashlib.blake2s(seal_raw).digest())
    elif hasattr(hub, f'takara.seal.{seal}.derive'):
        key = getattr(hub, f'takara.seal.{seal}.derive')(seal_raw)
    else:
        key = seal_raw
    hub.takara.CRYPT[unit] = {}
    hub.takara.CRYPT[unit]['crypter'] = cryptography.fernet.Fernet(key)


async def encrypt(hub, unit, data):
    '''
    Using the key from the given unit encrypt the raw bytes found in data
    and return the encryption string
    '''
    crypter = hub.takara.CRYPT[unit]['crypter']
    return crypter.encrypt(data)


async def decrypt(hub, unit, data):
    '''
    Using the key from the given unit decrypt the raw bytes found in data
    and return the clear string
    '''
    crypter = hub.takara.CRYPT[unit]['crypter']
    return crypter.decrypt(data)




'''
Create a unit storage system based on a local file directory
'''
# Import python libs
import os
import shutil
# Import third party libs
import msgpack
# Import local libs
import takara.exc

__virtualname__ = 'file'
__func_alias__ = {'set_': 'set', 'list_': 'list'}
UNIT_FN = 'config.mp'


async def config(hub, data_dir):
    '''
    Get the existing configuration for all available unit files
    '''
    ret = {}
    data_dir = data_dir
    for unit_name in os.listdir(data_dir):
        unit_file = os.path.join(data_dir, unit_name, UNIT_FN)
        if os.path.isfile(unit_file):
            with open(unit_file, 'rb') as rfh:
                ret[unit_name] = msgpack.loads(rfh.read(), raw=False)
    hub.takara.UNITS.update(ret)


async def create(hub, **kw):
    '''
    Given a directory to create the store in, make the root directory used
    for the store adn place the unit confiruration inside.
    '''
    data_dir = kw['data_dir']
    unit_name = kw['unit']
    unit_root = os.path.join(kw['unit_dir'], unit_name)
    cipher = kw['cipher']
    seal = kw['seal']
    seal_data = kw['seal_data']
    unit_dir = os.path.join(data_dir, unit_name)
    unit_file = os.path.join(unit_dir, UNIT_FN)
    if os.path.isfile(unit_file):
        raise takara.exc.UnitExistsError(f'Unit "{unit_name}" exists and is defined in file "{unit_file}", either delete this unit or chose another name')
    if not os.path.isdir(unit_dir):
        os.makedirs(unit_dir)
    if not os.path.isdir(unit_root):
        os.makedirs(unit_root)
    data = {'cipher': cipher,
            'seal': seal,
            'unit_root': unit_root,
            'seal_data': seal_data,
            'store': 'file',
            'unit_file': unit_file,
            }
    old_umask = os.umask(54)
    with open(unit_file, 'wb+') as wfh:
        wfh.write(msgpack.dumps(data, use_bin_type=True))
    os.umask(old_umask)
    # refresh the unit config on the hub
    await hub.takara.store.file.config(data_dir=data_dir)


async def set_(hub, unit, path, value):
    '''
    Given the named unit, set the named value. The value should already be
    encrypted using the named unit's sealing mechanism
    '''
    if unit not in hub.takara.UNITS:
        raise takara.exc.UnitMissingError('The named unit "{unit}" is not available')
    unit_config = hub.takara.UNITS[unit]
    unit_root = unit_config['unit_root']
    fn_path = os.path.join(unit_root, path.replace('.', os.sep))
    fn_dir = os.path.dirname(fn_path)
    if not os.path.isdir(fn_dir):
        os.makedirs(fn_dir)
    old_umask = os.umask(54)
    with open(fn_path, 'wb+') as wfh:
        wfh.write(value)
    os.umask(old_umask)


async def get(hub, unit, path):
    '''
    Given the unit name and the path, get the encrypted data from the given path
    '''
    if unit not in hub.takara.UNITS:
        raise takara.exc.UnitMissingError('The named unit "{unit}" is not available')
    unit_config = hub.takara.UNITS[unit]
    unit_root = unit_config['unit_root']
    fn_path = os.path.join(unit_root, path.replace('/', os.sep))
    if not os.path.isfile(fn_path):
        raise takara.exc.PathMissingError('The named path "{path}" is not present in unit "{unit}"')
    with open(fn_path, 'rb') as rfh:
        return rfh.read()


async def rm(hub, unit, path):
    '''
    Remove the given path from the system
    '''
    if unit not in hub.takara.UNITS:
        raise takara.exc.UnitMissingError('The named unit "{unit}" is not available')
    unit_config = hub.takara.UNITS[unit]
    unit_root = unit_config['unit_root']
    fn_path = os.path.join(unit_root, path.replace('/', os.sep))
    if os.path.isfile(fn_path):
        os.remove(fn_path)
    elif os.path.isdir(fn_path):
        shutil.rmtree(fn_path)


async def list_(hub, unit, path='/'):
    '''
    List all of the values in a given path location
    '''
    ret = []
    if unit not in hub.takara.UNITS:
        raise takara.exc.UnitMissingError('The named unit "{unit}" is not available')
    unit_config = hub.takara.UNITS[unit]
    unit_root = unit_config['unit_root']
    fn_path = os.path.join(unit_root, path.replace('/', os.sep))
    for root, dirs, files in os.walk(fn_path):
        rel_root = os.relpath(root, fn_path)
        for fn in files:
            ret.append(os.path.join(rel_root, fn))
    return ret


async def rm_unit(hub, unit):
    '''
    Remove the named unit from the system
    '''
    if unit not in hub.takara.UNITS:
        raise takara.exc.UnitMissingError('The named unit "{unit}" is not available')
    unit_config = hub.takara.UNITS.pop(unit)
    shutil.rmtree(unit_config['unit_root'])
    os.remove(unit_config['unit_file'])


async def rename_unit(hub, unit, new_unit):
    '''
    Take the given unit and rename it
    '''
    if unit not in hub.takara.UNITS:
        raise takara.exc.UnitMissingError('The named unit "{unit}" is not available')
    if new_unit in hub.takara.UNITS:
        raise takara.exc.UnitExistsError(f'Unit "{new_unit}" exists, please define another name')
    unit_config = hub.takara.UNITS.pop(unit)
    unit_root = unit_config['unit_root']
    unit_dir = os.path.dirname(unit_root)
    unit_tgt = os.path.join(unit_dir, new_unit)
    unit_file = unit_config['unit_file']
    unit_data_dir = os.path.dirname(unit_file)
    data_dir = os.path.dirname(os.path.dirname(unit_file))
    new_data_dir = os.path.join(data_dir, new_unit)
    new_unit_file = os.path.join(new_data_dir, UNIT_FN)
    if not os.path.isdir(new_data_dir):
        os.makedirs(new_data_dir)
    unit_config['unit_root'] = unit_tgt
    shutil.move(unit_root, unit_tgt)
    with open(new_unit_file, 'wb+') as wfh:
        wfh.write(msgpack.dumps(unit_config, use_bin_type=True))
    await hub.takara.store.file.config(data_dir=data_dir)
    shutil.rmtree(unit_data_dir)



'''
The most simple seal, a password. The password is stored in a salted hash in
the unit config. The password itself is used to create a seperate key for the
given cipher.

This implementation uses PBKDF2, this is not nessisarily the best choice, but
it is highly portable, well supported and has a great track record. It is
also natively available inside of the standard python library.

If other password implementations are desired we plan to add vertical app
merged systems to support argon2, bcrypt, and scrypt.
'''
# Import python libs
import hashlib
import binascii
import random
import time
import os
from getpass import getpass

# Don't echo the password back to the user
EXPOSE_SEAL_RAW = False
ENCODE_RAW = True


async def gen(hub, seal_raw=None, cipher=None):
    '''
    Generate a password, in this case, gather the password from the cli unless
    a password is provided
    '''
    if seal_raw is not None:
        return seal_raw
    return getpass()


async def create(hub, **kw):
    '''
    Given a password, create the salted hash to store for verification. This
    function is not unit specific but it intended to be called before the
    unit itself is created with the result passed down to the unit creation.
    '''
    algo = kw.get('passwd_hash_algo', 'sha512')
    iterations = kw.get('passwd_iterations', 100000)
    passwd = kw['seal_raw'].encode('utf-8')
    salt = os.urandom(1024)
    raw = hashlib.pbkdf2_hmac(algo, passwd, salt, iterations)
    return b':'.join([
        algo.encode('utf-8'),
        binascii.hexlify(raw),
        binascii.hexlify(salt),
        str(iterations).encode('utf-8')
        ])


async def verify(hub, passwd, phash):
    '''
    Verify the given phash and password combination
    '''
    passwd = passwd.encode('utf-8')
    phash = phash
    comps = phash.split(b':')
    if len(comps) != 4:
        return False
    algo = comps[0].decode('utf-8')
    raw = binascii.unhexlify(comps[1])
    salt = binascii.unhexlify(comps[2])
    iterations = int(comps[3])
    new_raw = hashlib.pbkdf2_hmac(algo, passwd, salt, iterations)
    good = True
    for idx, part in enumerate(raw):
        if part != new_raw[idx]:
            good = False
    if not good:
        base = 0.2
        base = base + random.randint(1000, 9000) * 0.0001
        time.sleep(base)
    return good



__virtualname__ = 'shamir'
# Import python libs
import os
import codecs
import base64
import hashlib
from getpass import getpass

# Import third party libs
import cryptography.fernet

# Tell takara to present the seal_raw to the end user, as the seal can only
# be generated
EXPOSE_SEAL_RAW = True

# Computation variables
SECRET_LEN = 32
POLY = 115792089237316195423570985008687907853269984665640564039457584007913129640997
CHECK = b'Oh king eh? Very nice!'


async def get(hub, seal_raw=None, cipher=None):
    if seal_raw is not None:
        return seal_raw
    first = getpass('Please enter the first of three keys: ')
    second = getpass('Please enter the second of three keys: ')
    third = getpass('Please enter the third of three keys: ')
    return '{first}:{second}:{third}'


async def gen(hub, seal_raw=None, cipher=None):
    if seal_raw is not None:
        return seal_raw
    key = getattr(hub, f'takara.cipher.{cipher}.gen')()
    comps = split(key, 5, 3)
    ret = ''
    for comp in comps:
        ret += f'{comp[0]}|{comp[1].hex()}:'
    return ret[:-1]


async def create(hub, **kw):
    comps = _to_comps(kw['seal_raw'].encode('utf-8'))
    secret = combine(3, comps)
    key = base64.urlsafe_b64encode(hashlib.blake2s(secret).digest())
    crypter = cryptography.fernet.Fernet(key)
    return crypter.encrypt(CHECK)


async def verify(hub, passwd, phash):
    '''
    Verify that the key derived from the source is good
    '''
    comps = _to_comps(passwd.encode('utf-8'))
    secret = combine(3, comps)
    key = base64.urlsafe_b64encode(hashlib.blake2s(secret).digest())
    crypter = cryptography.fernet.Fernet(key)
    if crypter.decrypt(phash) == CHECK:
        return True
    return False


def derive(hub, seal_raw):
    '''
    Derive the correct key from the seal_raw inputs
    '''
    comps = _to_comps(seal_raw)
    secret = combine(3, comps)
    key = base64.urlsafe_b64encode(hashlib.blake2s(secret).digest())
    return key


def _to_comps(seal_raw):
    ret = []
    for comp in seal_raw.split(b':'):
        num, hex_digest = comp.split(b'|')
        ret.append((int(num), bytes.fromhex(hex_digest.decode())))
    return ret


def hexdec(data):
    return codecs.getdecoder('hex')(data)[0]


def hexenc(data):
    return codecs.getencoder('hex')(data)[0].decode('utf-8')


def bytes_to_long(raw):
    return int(hexenc(raw), 16)


def long_to_bytes(n, size=32):
    res = hex(int(n))[2:].rstrip("L")
    if len(res) % 2 != 0:
        res = "0" + res
    s = hexdec(res)
    if len(s) != size:
        s = (size - len(s)) * b"\x00" + s
    return s

def _lshift(x, bits):
    return x * (1 << bits)


def _field_mult(x, y):
    b = x
    z = b if y & 1 == 1 else 0
    for i in range(1, SECRET_LEN * 8):
        b = _lshift(b, 1)
        if (b >> (SECRET_LEN * 8)) & 1 == 1:
            b ^= POLY
        if y & (1 << i):
            z ^= b
    return z


def _horner(t, x, coef):
    y = coef[t - 1]
    for i in range(t - 1, 0, -1):
        y = _field_mult(y, x)
        y ^= coef[i - 1]
    return y


def _field_invert(x):
    u, v = x, POLY
    g, z = 0, 1
    while u > 1:
        i = len(bin(u)[2:]) - len(bin(v)[2:])
        if i < 0:
            u, v = v, u
            z, g = g, z
            i = -i
        u = u ^ _lshift(v, i)
        z = z ^ _lshift(g, i)
    return z


def _calculate_li0(t, x, i):
    li0 = 1
    for j in range(t):
        if j == i:
            continue
        li0 = _field_mult(li0, x[j])
        li0 = _field_mult(li0, _field_invert(x[i] ^ x[j]))
    return li0


def split(secret, n, t):
    '''
    Split the given secret, n being the number of required keys
    and t being the total number of keys
    '''
    coef = [bytes_to_long(secret[::-1])]
    if n < 0 or t < 0 or n < t or not secret:
        raise ValueError('Invalid parameters specified')
    for i in range(1, t):
        coef.append(bytes_to_long(os.urandom(SECRET_LEN)))
    out = []
    for i in range(1, n + 1):
        out.append((i, long_to_bytes(_horner(t, i, coef))[::-1]))
    return out


def combine(t, parts):
    '''
    Combine the secret from the parts given
    '''
    if t <= 0 or not parts:
        raise ValueError('Invalid parameters specified')
    if len(parts) != len(set(s[1] for s in parts)):
        raise ValueError('Equal parts found')
    x, y = list(zip(*[(s[0], bytes_to_long(s[1][::-1])) for s in parts]))
    sec = 0
    for i in range(t):
        li0 = _calculate_li0(t, x, i)
        li0si = _field_mult(li0, y[i])
        sec = sec ^ li0si
    return long_to_bytes(sec)[::-1]



class TakaraBaseException(Exception):
    '''
    Base exception where all of Takara's exceptions derive
    '''


class TakaraError(TakaraBaseException):
    '''
    General purpose Takara exception to signal an error
    '''


class UnitExistsError(TakaraBaseException):
    '''
    Raised when trying to create a unit that is already present
    '''


class UnitMissingError(TakaraBaseException):
    '''
    Raised when trying to access a unit that is not present
    '''


class PathMissingError(TakaraBaseException):
    '''
    Raised when trying to access an unavailiable path
    '''


class UnsealError(TakaraBaseException):
    '''
    Raised when the system fails to unseal a unit
    '''



'''
Functions used to migrate a unit from one store/seal/cipher to another
'''

async def unit(hub, unit, new_unit, **kw):
    '''
    Migrate the named unit to a new store/seal/cipher
    '''
    if unit not in hub.takara.UNITS:
        raise takara.exc.UnitMissingError('The named unit "{unit}" is not available')
    unit_config = hub.takara.UNITS[unit]
    if kw.get('store', unit_config['store']) == unit_config['store']:
        # Unit stores will collide! Run rename
        await getattr(hub, f'takara.store.{kw["store"]}.rename_unit')(unit, new_unit)


async def seal(hub, unit, seal, seal_raw):
    '''
    Migrate the given unit to a new seal type. This will require re-encrypting
    all data against the new seal.
    '''
    if unit not in hub.takara.UNITS:
        raise takara.exc.UnitMissingError('The named unit "{unit}" is not available')
    unit_config = hub.takara.UNITS[unit]
    tmp_unit = os.urandom(16).hex()
    unit_config['unit'] = tmp_unit
    unit_config['seal'] = seal
    unit_config['seal_raw'] = seal_raw
    strore = unit_config['store']
    await getattr(hub, f'takara.store.{store}.create')(**unit_config)
    keys = await getattr(hub, f'takara.store.{store}.list')(unit)
    for key in keys:
        orig = await getattr(hub, f'takara.store.{store}.get')(unit, path)
        await getattr(hub, f'takara.store.{store}.set')(tmp_unit, path, orig)
    await getattr(hub, f'takara.store.{store}.rm_unit')(unit)
    await getattr(hub, f'takara.store.{store}.rename_unit')(tmp_unit, unit)
    # RM original unit
    # Move tmp unit to orig name
